from __future__ import print_function, unicode_literals
__all__ = ['bpch1', 'ncf2bpch', 'bpch_base']
import os

# part of the default Python distribution
from collections import OrderedDict
from warnings import warn
import unittest

# numpy is a very common installed library
import numpy as np
from numpy import fromfile, memmap, dtype, arange, zeros, diff, concatenate
from numpy import append, pi, sin

from PseudoNetCDF.sci_var import PseudoNetCDFVariable, PseudoNetCDFFile
from PseudoNetCDF._getwriter import registerwriter


class OrderedDefaultDict(OrderedDict):
    def __init__(self, *args, **kwargs):
        if not args:
            self.default_factory = None
        else:
            if not (args[0] is None or callable(args[0])):
                raise TypeError('first argument must be callable or None')
            self.default_factory = args[0]
            args = args[1:]
        super(OrderedDefaultDict, self).__init__(*args, **kwargs)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = default = self.default_factory()
        return default


# These variables define the binary format of the header blocks
# and are only for internal
_general_header_type = dtype('>i4, S40, >i4, >i4, S80, >i4')
_datablock_header_type = dtype('>i4, S20, 2>f4, >i4, >i4, >i4, >i4, S40, ' +
                               '>i4, S40, >f8, >f8, S40, 3>i4, 3>i4, >i4, >i4')
_first_header_size = (_general_header_type.itemsize +
                      _datablock_header_type.itemsize)


class defaultdictfromkey(OrderedDefaultDict):
    """
    defaultdictfromkey dynamically produces dictionary items
    for keys, using the default_factor function called
    with the key as an argument
    """

    def __missing__(self, key):
        """
        __missing__(key) # Called by __getitem__ for missing key; pseudo-code:
        if self.default_factory is None: raise KeyError((key,))
        self[key] = value = self.default_factory()
        return value
        """
        return self.default_factory(key)


class defaultdictfromthesekeys(OrderedDefaultDict):
    """
    defaultdictfromthesekeys dynamically produces dictionary items
    for known keys, using the default_factor function called
    with the key as an argument
    """

    def __init__(self, keys, default_factory=None):
        """
        keys - iterable of keys that default_factory can produce
        default_factory - function that takes a key as an argument
                          to create a dictionary item
        """
        self._keys = [k for k in keys]
        OrderedDefaultDict.__init__(self, default_factory)

    def __iter__(self):
        """
        Just like dictionary, but iterates through pre-defined keys
        """
        for i in self._keys:
            yield i

    def values(self):
        """
        Just like dictionary, but iterates through pre-defined key values
        """
        for k in self.keys():
            yield self[k]

    def items(self):
        """
        Just like dictionary, but iterates through pre-defined keys and their
        calculated values
        """
        for k in self.keys():
            yield (k, self[k])

    def keys(self):
        """
        Just like dictionary, but iterates through pre-defined keys
        """
        return [k for k in self]

    def __setitem__(self, key, value):
        """
        Add value with key and add to pre-defined keys
        """
        val = OrderedDefaultDict.__setitem__(self, key, value)
        if key not in self._keys:
            self._keys.append(key)
        return val

    def __delitem__(self, key):
        """
        Delete value with key and remove pre-defined key
        """
        if key in self._keys:
            ki = self._keys.index(key)
            del self._keys[ki]
        return OrderedDefaultDict.__delitem__(self, key)

    def pop(self, key):
        """
        Pop value with key and remove pre-defined key
        """
        val = OrderedDefaultDict.pop(self, key)
        if key in self._keys:
            ki = self._keys.index(key)
            del self._keys[ki]
        return val

    def __missing__(self, key):
        """
        __missing__(key) # Called by __getitem__ for missing key; pseudo-code:
        if self.default_factory is None: raise KeyError((key,))
        self[key] = value = self.default_factory()
        return value
        """
        if key in self._keys:
            return self.default_factory(key)
        else:
            raise KeyError("%s not found" % (key, ))


class defaultpseudonetcdfvariable(defaultdictfromthesekeys):
    """
    Overwrites __repr__ function to show variables
    """

    def __contains__(self, key):
        return key in self._keys

    def __repr__(self):
        out = "{"
        for k in self._keys:
            out += "\n  '%s': PseudoNetCDFVariable(...)" % k
        out += "\n}"
        return out

    def __str__(self):
        return self.__repr__()


class _diag_group(PseudoNetCDFFile):
    """
    This object acts as a PseudoNetCDF file that gets data from the parent
    object.
    """

    def __init__(self, parent, groupname, groupvariables):
        """
        parent - a PseudoNetCDFFile
        groupname - a string describing the group
        """
        template = '%s_%%s' % groupname

        def getvar(key):
            try:
                return parent.variables[template % key]
            except (KeyError, ValueError):
                return parent.variables[key]
        self._parent = parent
        if 'BXHGHT-$_BXHEIGHT' not in groupvariables:
            mymetakeys = [k for k in metakeys if k != 'VOL']
        else:
            mymetakeys = metakeys
        self.variables = defaultpseudonetcdfvariable(
            list(groupvariables) + mymetakeys, getvar)

    def __getattr__(self, key):
        try:
            return object.__getattr__(self, key)
        except AttributeError as e:
            if key != 'groups':
                return getattr(self._parent, key)
            else:
                raise e

# This class is designed to operate like a dictionary, but
# dynamically create variables to return to the user


class _tracer_lookup(defaultpseudonetcdfvariable):
    """
    _tracer_lookup: finds geos_chem tracer indices from names and returns
                    netcdf like variable
    """

    def __init__(self, parent, datamap, tracerinfo, diaginfo, keys,
                 noscale=False, nogroup=False):
        """
        parent: NetCDF-like object to serve dimensions
        datamap: array of pre-dimensioned and datatyped values
                 dim(tstep, i, j, k)
        tracerinfo: dictionary of tracer data keyed by ordinal
        keys: list of keys to serve
        """
        self.noscale = noscale
        self.nogroup = nogroup
        self._tracer_data = tracerinfo
        self._diag_data = diaginfo
        self._memmap = datamap
        self._parent = parent
        self._special_keys = set(metakeys)
        self._keys = keys + [k for k in self._special_keys if k not in keys]
        if 'BXHGHT-$_BXHEIGHT' not in keys:
            ki = self._keys.index('VOL')
            del self._keys[ki]
        self._example_key = keys[0]

    def __missing__(self, key):
        from ._vertcoord import geos_etai_pressure, geos_etam_pressure
        from ._vertcoord import geos_hyam, geos_hybm
        if key in ('latitude', 'latitude_bounds'):
            yres = self._parent.modelres[1]
            if self._parent.halfpolar == 1:
                data = concatenate(
                    [[-90.], arange(-90. + yres / 2., 90., yres), [90.]])
            else:
                data = arange(-90, 90 + yres, yres)

            dims = ('latitude',)
            dtype = 'f'
            kwds = dict(standard_name="latitude", long_name="latitude",
                        units="degrees_north", base_units="degrees_north",
                        axis="Y")
            if key == 'latitude':
                data = data[:-1] + diff(data) / 2.
                kwds['bounds'] = 'latitude_bounds'
            else:
                dims += ('nv',)
                data = data.repeat(2, 0)[1:-1].reshape(-1, 2)
            example = self[self._example_key]
            sj = getattr(example, 'STARTJ', 0)
            data = data[sj:sj + example.shape[2]]
        elif key in ('longitude', 'longitude_bounds'):
            xres = self._parent.modelres[0]
            i = arange(0, 360 + xres, xres)
            data = i - (180 + xres / 2. * self._parent.center180)
            dims = ('longitude',)
            dtype = 'f'
            kwds = dict(standard_name="longitude", long_name="longitude",
                        units="degrees_east", base_units="degrees_east",
                        axis="X")
            if key == 'longitude':
                data = data[:-1] + diff(data) / 2.
                kwds['bounds'] = 'longitude_bounds'
            else:
                dims += ('nv',)
                data = data.repeat(2, 0)[1:-1].reshape(-1, 2)
            example = self[self._example_key]
            si = getattr(example, 'STARTI', 0)
            data = data[si:si + example.shape[3]]
        elif key == 'AREA':
            lon = self['longitude']
            xres = self._parent.modelres[0]
            nlon = 360. / xres
            latb = self['latitude_bounds']
            Re = self['crs'].semi_major_axis
            latb = append(latb[:, 0], latb[-1, 1])
            latr = pi / 180. * latb
            data = 2. * pi * Re * Re / \
                (nlon) * (sin(latr[1:]) - sin(latr[:-1]))
            data = data[:, None].repeat(lon.size, 1)
            kwds = dict(units='m**2', base_units='m**2', grid_mapping="crs")
            dtype = 'f'
            dims = ('latitude', 'longitude')
        elif key == 'VOL':
            try:
                bxhght = self['BXHGHT-$_BXHEIGHT']
                area = self['AREA']
            except KeyError:
                raise KeyError(
                    'Volume is only available if BXHGHT-$_BXHEIGHT was output')
            data = area[None, None] * bxhght
            kwds = dict(units='m**3', base_units='m**3', grid_mapping="crs")
            dtype = 'f'
            dims = ('time', 'layer', 'latitude', 'longitude')
            if len(['layer' in dk_ for dk_ in self._parent.dimensions]) > 1:
                dims = ('time', 'layer%d' %
                        data.shape[1], 'latitude', 'longitude')
        elif key == 'crs':
            dims = ()
            kwds = OrderedDict(grid_mapping_name="latitude_longitude",
                               semi_major_axis=6375000.,
                               semi_minor_axis=6375000.)
            dtype = 'i'
            data = np.array(0, dtype=dtype)
        elif key in ('time', 'time_bounds'):
            tmp_key = self._example_key
            data = np.array([self['tau0'], self['tau1']]).T
            dims = ('time', 'tnv')
            if key == 'time':
                data = data.mean(1)
                dims = ('time',)

            dtype = 'd'
            kwds = dict(units='hours since 1985-01-01 00:00:00 UTC',
                        base_units='hours since 1985-01-01 00:00:00 UTC',
                        standard_name=key, long_name=key, var_desc=key)
            if key == 'time':
                kwds['bounds'] = 'time_bounds'
        elif key == 'hyai':
            data = self._parent.Ap
            dims = ('layer_bounds', )
            dtype = data.dtype.char
            if dims[0] not in self._parent.dimensions:
                self._parent.createDimension(dims[0], data.size)
            kwds = dict(units="hPa",
                        long_name="hybrid A coefficient at layer interfaces",
                        note="unit consistent with GEOS-Chem pressure outputs")
        elif key == 'hyam':
            data = geos_hyam[self._parent.vertgrid]
            dims = ('layer', )
            dtype = data.dtype.char
            if dims[0] not in self._parent.dimensions:
                self._parent.createDimension(dims[0], data.size)
            kwds = dict(units="hPa",
                        long_name="hybrid B coefficient at layer midpoints",
                        note="unit consistent with GEOS-Chem pressure outputs")
        elif key == 'hybi':
            data = self._parent.Bp
            dims = ('layer_bounds',)
            dtype = data.dtype.char
            if dims[0] not in self._parent.dimensions:
                self._parent.createDimension(dims[0], data.size)
            kwds = dict(units="1",
                        long_name="hybrid B coefficient at layer interfaces")
        elif key == 'hybm':
            data = geos_hybm[self._parent.vertgrid]
            dims = ('layer', )
            dtype = data.dtype.char
            if dims[0] not in self._parent.dimensions:
                self._parent.createDimension(dims[0], data.size)
            kwds = dict(
                units="1", long_name="hybrid B coefficient at layer midpoints")
        elif key[:5] == 'layer':
            nlays = len(self._parent.dimensions['layer'])
            nedges = nlays + 1
            if key[5:] in (str(nedges), '_edges', '_bounds'):
                data = np.arange(0, nedges, dtype='f')
            else:
                data = np.arange(1, nedges, dtype='f')

            data = data[:len(self._parent.dimensions[key])]
            dims = (key,)
            dtype = 'f'
            kwds = dict(units='level', base_units='model layer',
                        standard_name='model layer', long_name=key,
                        var_desc=key, axis="Z")
            kwds = dict(standard_name="hybrid_sigma_pressure",
                        long_name="hybrid level at layer midpoints",
                        units="level",
                        positive="up",)
        elif key in ('etai_pressure', 'etam_pressure'):
            nlays = len(self._parent.dimensions['layer'])
            nedges = nlays + 1
            if key[:4] == 'etai':
                data = geos_etai_pressure[self._parent.vertgrid]
            else:
                data = geos_etam_pressure[self._parent.vertgrid]

            data = data[:nedges]
            if 'etai' in key:
                dims = ('layer_bounds',)
            else:
                dims = ('layer',)
            dtype = 'f'
            kwds = dict(units='hPa', base_units='hPa',
                        standard_name=('atmosphere_hybrid_sigma_pressure_' +
                                       'coordinate'),
                        long_name=key, var_desc=key)
        elif key == 'tau0':
            tmp_key = self._example_key
            data = self._memmap[tmp_key]['header']['f10']
            dims = ('time',)
            dtype = 'd'
            kwds = dict(units='hours since 1985-01-01 00:00:00 UTC',
                        base_units='hours since 1985-01-01 00:00:00 UTC',
                        standard_name=key, long_name=key, var_desc=key)
        elif key == 'tau1':
            tmp_key = self._example_key
            data = self._memmap[tmp_key]['header']['f11']
            dims = ('time',)
            dtype = 'd'
            kwds = dict(units='hours since 1985-01-01 00:00:00 UTC',
                        base_units='hours since 1985-01-01 00:00:00 UTC',
                        standard_name=key, long_name=key, var_desc=key)
        else:
            dtype = 'f'
            key = str(key)
            header = self._memmap[key]['header'][0]
            sl, sj, si = header['f14'][::-1] - 1
            group = header['f7'].strip().decode('ASCII')
            offset = self._diag_data.get(group, {}).get('offset', 0)
            tracerid = header['f8']
            ord = header['f8'] + offset
            base_units = header['f9']
            reserved = header['f12']
            scale = self._tracer_data[ord]['SCALE']
            molwt = self._tracer_data[ord]['MOLWT']
            carbon = self._tracer_data[ord]['C']
            units = self._tracer_data[ord]['UNIT']
            tmp_data = self._memmap[key]['data']
            dims = ('time', 'layer', 'latitude', 'longitude')
            if len(['layer' in dk_ for dk_ in self._parent.dimensions]) > 1:
                dims = ('time', 'layer%d' %
                        tmp_data.dtype['f1'].shape[0], 'latitude', 'longitude')
            kwds = dict(scale=scale, kgpermole=molwt, carbon=carbon,
                        units=units, base_units=base_units, standard_name=key,
                        long_name=key, var_desc=key,
                        coordinates=' '.join(dims), grid_mapping="crs",
                        reserved=reserved, tracerid=tracerid, category=group)

            try:
                assert((tmp_data['f0'] == tmp_data['f2']).all())
            except Exception:
                raise ValueError('Could not parse with bpch; try bpch2')

            if self.noscale:
                data = tmp_data['f1']
            else:
                data = tmp_data['f1'] * scale
            if any([sl != 0, sj != 0, si != 0]):
                nl, nj, ni = header['f13'][::-1]
                kwds['STARTI'] = si
                kwds['STARTJ'] = sj
                kwds['STARTK'] = sl
        return PseudoNetCDFVariable(self._parent, key, dtype, dims,
                                    values=data, **kwds)


coordkeys = ['time', 'latitude', 'longitude', 'layer',
             'time_bounds', 'layer_bounds',
             'latitude_bounds', 'longitude_bounds', 'crs']
metakeys = ['VOL', 'AREA', 'tau0', 'tau1',
            'etam_pressure', 'etai_pressure', 'hyam', 'hybm',
            'hyai', 'hybi'] + coordkeys


class bpch_base(PseudoNetCDFFile):
    """
    Binary Punch File reader
    """

    def getTimes(self, bounds=False):
        from datetime import timedelta
        from ..coordutil import _parse_ref_date
        if bounds:
            timeb = self.variables['time_bounds']
            timeunit = timeb.units.strip()
            time = np.append(timeb[:, 0], timeb[-1, -1])
        else:
            time = self.variables['time']
            timeunit = time.units.strip()
        if 'since' in timeunit:
            unit, base = timeunit.split(' since ')
            sdate = _parse_ref_date(base)
            out = sdate + \
                np.array([timedelta(**{unit: float(i)}) for i in time[:]])
            return out
        else:
            return time

    def plot(self, varkey, plottype='longitude-latitude', ax_kw=None,
             plot_kw=None, cbar_kw=None, map_kw=None, dimreduction='mean',
             laykey=None):
        """
        Parameters
        ----------
        self : the bpch file
        varkey : the variable to plot
        plottype : longitude-latitude, latitude-pressure, longitude-pressure,
                   vertical-profile, time-longitude, time-latitude,
                   time-pressure, default, longitude-latitude
        ax_kw : keywords for the axes to be created
        plot_kw : keywords for the plot (plot, scatter, or pcolormesh) to be
                  created
        cbar_kw : dictionary or bool or None
            keywords for the colorbar; if True or None, use defaults.
            If False, do not create a colorbar
        map_kw : dictionary or bool or None
            keywords for the getMap routine, which is only used with
            map capable dimensions (ie, plottype='longitude-latitude')
            If True or None, use default configuration dict(countries=True,
            coastlines=True, states=False, counties=False). If False,
            do not draw a map.
        dimreduction : dimensions not being used in the plot are removed
                       using applyAlongDimensions(dimkey=dimreduction) where
                       each dimenions
        laykey : name of the layer dimension (e.g., layer47)
        """
        import matplotlib.pyplot as plt
        from ..coordutil import getbounds

        if ax_kw is None:
            ax_kw = {}

        if plot_kw is None:
            plot_kw = {}

        if cbar_kw is None or cbar_kw is True:
            cbar_kw = {}

        if map_kw is None or map_kw is True:
            map_kw = {}

        apply2dim = {}
        var = self.variables[varkey]
        varunit = varunit = varkey + ' (' + getattr(var, 'units', None) + ')'
        dimlens = dict([(dk, len(self.dimensions[dk]))
                        for dk in var.dimensions])
        dimpos = dict([(dk, di) for di, dk in enumerate(var.dimensions)])
        if laykey is None:
            for k in list(dimlens):
                if k.startswith('layer'):
                    laykey = k
                    break
        xkey, ykey = plottype.split('-')
        for dimkey in 'longitude latitude time'.split():
            if dimkey in var.dimensions:
                if dimkey not in (xkey, ykey) and dimlens.get(dimkey, 1) > 1:
                    apply2dim[dimkey] = dimreduction

        if 'pressure' not in (xkey, ykey) and dimlens.get(laykey, 1) > 1:
            apply2dim[laykey] = dimreduction

        if len(apply2dim) > 0:
            myf = self.applyAlongDimensions(**apply2dim)
            var = myf.variables[varkey]
            dimlens = dict([(dk, len(self.dimensions[dk]))
                            for dk in var.dimensions])
        else:
            myf = self
        if ykey in ('profile',):
            if xkey == 'pressure':
                vaxi = var.dimensions.index(laykey)
            else:
                vaxi = var.dimensions.index(xkey)
            vsize = var.shape[vaxi]
            vals = np.rollaxis(var[:], vaxi).reshape(vsize, -1)
        else:
            vals = var[:].squeeze()

        ax = plt.gca(**ax_kw)
        if ykey in ('profile',):
            if xkey == 'pressure':
                y = myf.variables['etam_pressure']
            else:
                y = getbounds(myf, xkey)

            x0 = vals[:].min(0)
            xm = vals[:].mean(0)
            x1 = vals[:].max(0)
            ax.fill_betweenx(y=y, x0=x0, x1=x1, label=varkey + '(min, max)')
            ax.plot(xm, y, label=varkey, **plot_kw)
            ax.set_ylabel(xkey)
            ax.set_xlabel(varunit)
            return ax

        if xkey == 'time':
            x = myf.getTimes(bounds=True)
            x = plt.matplotlib.dates.date2num(x)
        elif xkey == 'pressure':
            x = myf.variables['etai_pressure'][:dimlens[laykey] + 1]
        else:
            x = getbounds(myf, xkey)
        if ykey == 'time':
            y = myf.getTimes(bounds=True)
            y = plt.matplotlib.dates.date2num(y)
        elif ykey == 'pressure':
            y = myf.variables['etai_pressure'][:dimlens[laykey] + 1]
        else:
            y = getbounds(myf, ykey)

        if dimpos[xkey] < dimpos[ykey]:
            vals = vals.T
        p = ax.pcolormesh(x, y, vals, **plot_kw)
        if cbar_kw is not False:
            ax.figure.colorbar(p, label=varunit, **cbar_kw)
        if ykey == 'pressure':
            ax.set_ylim(y.max(), y.min())
        dfmt = plt.matplotlib.dates.AutoDateFormatter(
            plt.matplotlib.dates.AutoDateLocator(interval_multiples=True))
        dfmt.scaled[365] = '%F'
        dfmt.scaled[30] = '%F'
        if xkey == 'time':
            ax.xaxis.set_major_formatter(dfmt)
        if ykey == 'time':
            ax.yaxis.set_major_formatter(dfmt)
        if plottype == 'longitude-latitude' and map_kw is not False:
            try:
                bmap = myf.getMap(**map_kw)
                bmap.drawcoastlines(ax=ax)
                bmap.drawcountries(ax=ax)
            except Exception:
                pass
        else:
            ax.set_xlabel(xkey)
            ax.set_ylabel(ykey)
        return ax

    def interpSigma(self, vglvls, vgtop=None, interptype='linear',
                    extrapolate=False, fill_value='extrapolate',
                    layerdims=None,
                    approach='eta', verbose=0):
        """
        Parameters
        ----------
        self : the file to interpolate from must have VGLVLS
        vglvls : the new vglvls (edges)
        vgtop : Converting to new vgtop (Pascals)
        interptype : 'linear' or 'conserve'
             linear : uses a linear interpolation
             conserve : uses a mass conserving interpolation
        extrapolate : allow extrapolation beyond bounds with linear,
                      default False
        fill_value : set fill value (e.g, nan) to prevent extrapolation or edge
                     continuation
        layerdims : specify layer dimension, None will apply to all dimensions
                    named layer*
        approach :
             eta : use simple eta coordinates to calculate sigma and
                   interpolate
             pressure : requires surface pressure
        verbose : 0-inf show more

        Returns
        -------
        outf - ioapi_base PseudoNetCDFFile with al variables interpolated

        Notes
        -----
        When extrapolate is false, the edge values are used for points beyond
        the inputs.
        """
        etai = self.variables['etai_pressure'] * 100
        # grab input sigma coordinates
        myvglvls = (etai - vgtop) / (etai[0] - vgtop)

        # Use midpoint for sigmas in inputs and outputs
        zs = (myvglvls[:-1] + myvglvls[1:]) / 2.
        nzs = (vglvls[:-1] + vglvls[1:]) / 2.

        if interptype == 'linear':
            from ..coordutil import getinterpweights
            weights = getinterpweights(
                zs, nzs, kind=interptype, fill_value=fill_value,
                extrapolate=extrapolate)
            # Create a function for interpolation

            def interpsigma(data):
                if data.ndim == 1:
                    newdata = (weights[:data.shape[0]] * data[:, None]).sum(0)
                else:
                    newdata = (weights[None, :data.shape[0], :, None, None] *
                               data[:, :, None]).sum(1)
                return newdata
        else:
            raise ValueError(
                'interptype only implemented for "linear"; got ' + interptype
            )

        # Apply function on LAY
        if layerdims is None:
            layerdims = sorted([
                dk for dk in self.dimensions
                if (
                    dk.startswith('layer') and
                    not dk.endswith('_bounds') and
                    not dk == 'layer1')
            ])
            if verbose > 0:
                print(layerdims)
        dimfunc = dict([(layerkey, interpsigma) for layerkey in layerdims])
        outf = self.applyAlongDimensions(**dimfunc)

        return outf

    def subsetVariables(self, varkeys, inplace=False, keepcoords=True,
                        exclude=False):
        if keepcoords and not exclude:
            varkeys = ([key for key in coordkeys
                        if key not in varkeys and key in self.variables] +
                       varkeys)
        return PseudoNetCDFFile.subsetVariables(self, varkeys, inplace=inplace,
                                                exclude=exclude)

    def xy2ll(self, x, y):
        "see ll2xy"
        self.ll2xy(x, y)

    def ll2xy(self, lon, lat):
        """
        lon and lat are converted to local lon and lat (-180,180) (-90, 90)
        """
        lonr, latr = PseudoNetCDFFile.ll2xy(self, lon, lat)
        return np.degrees(lonr), np.degrees(latr)

    def ij2ll(self, i, j):
        """
        i, j to center lon, lat
        """
        lat = np.asarray(self.variables['latitude'])
        lon = np.asarray(self.variables['longitude'])
        return lon[i], lat[j]

    def ll2ij(self, lon, lat, bounds='error'):
        """
        Converts lon/lat to 0-based indicies (0,M), (0,N)

        Parameters
        ----------
        lon : scalar or iterable of longitudes in decimal degrees
        lat : scalar or iterable of latitudes in decimal degrees
        bounds : ignore, error, warn if i,j are out of domain

        Returns
        -------
        i, j : indices (0-based) for variables
        """
        late = np.asarray(self.variables['latitude_bounds'])
        lone = np.asarray(self.variables['longitude_bounds'])
        inlon = (lon >= lone[:, 0, None]) & (lon < lone[:, 1, None])
        inlat = (lat >= late[:, 0, None]) & (lat < late[:, 1, None])
        if not inlon.any(0).all() or not inlat.any(0).all():
            if bounds == 'error':
                raise ValueError('lat/lon not in domain')
            elif bounds == 'warn':
                warn('lat/lon not in domain')
        i = inlon.argmax(0)
        j = inlat.argmax(0)
        return i, j


class bpch1(bpch_base):
    """
    NetCDF-like class to interface with GEOS-Chem binary punch files

    f = bpch(path_to_binary_file)
    dim = f.dimensions[dkey] # e.g., dkey = 'longitude'

    # There are two ways to get variables.  Directly from
    # the file using the long name
    var = f.variables[vkey] # e.g., vkey = 'IJ-AVG-$_NOx'

    # Or through a group using the short name
    g = f.groups[gkey] # e.g., gkey = 'IJ-AVG-$'
    var = g.variables[vkey] # e.g., vkey = 'NOx'

    # The variable returned is the same either way
    print(f.dimensions)
    print(var.unit)
    print(var.dimensions)
    print(var.shape)

    """
    @classmethod
    def isMine(cls, path, *args, **kwds):
        try:
            # Read binary data for general header and first datablock header
            header_block = fromfile(
                path, dtype='bool', count=_first_header_size)

            # combine data for convenience
            ght = _general_header_type
            ghts = ght.itemsize
            dht = _datablock_header_type
            header = (tuple(header_block[:ghts].view(ght)[0]) +
                      tuple(header_block[ghts:].view(dht)[0]))

            # Verify that all Fortran unformatted buffers match
            try:
                assert(header[0] == header[2])
                assert(header[3] == header[5])
                return True
            except AssertionError:
                return False
        except Exception:
            return False

    @classmethod
    def from_ncf(cls, infile):
        outf = bpch_base()
        for pk in infile.ncattrs():
            pv = getattr(infile, pk)
            setattr(outf, pk, pv)

        for dk, dv in infile.dimensions.items():
            outf.copyDimension(dv, key=dk)

        for vk, vv in infile.variables.items():
            outf.copyVariable(vv, key=vk)

        return outf

    def _newlike(self):
        if isinstance(self, PseudoNetCDFFile):
            outt = bpch_base
            outf = outt.__new__(outt)
        else:
            outf = PseudoNetCDFFile()
        return outf

    def __init__(self, bpch_path, tracerinfo=None, diaginfo=None, mode='r',
                 timeslice=slice(None), noscale=False,
                 vertgrid='GEOS-5-REDUCED', nogroup=False):
        """
        bpch_path: path to binary punch file
        tracerinfo: path to ascii file with tracer definitions
        diaginfo: path to ascii file with diagnostic group definitions
        mode : {'r+', 'r', 'w+', 'c'}, optional
         |      The file is opened in this mode:
         |
         | +------+-----------------------------------------------------------+
         | | 'r'  | Open existing file for reading only.                      |
         | +------+-----------------------------------------------------------+
         | | 'r+' | Open existing file for reading and writing.               |
         | +------+-----------------------------------------------------------+
         | | 'w+' | Create or overwrite existing file for reading and writing.|
         | +------+-----------------------------------------------------------+
         | | 'c'  | Copy-on-write: assignments affect data in memory, but     |
         | |      | changes are not saved to disk.  The file on disk is       |
         | |      | read-only.                                                |
         | +------+-----------------------------------------------------------+
         timeslice: If the file is larger than 2GB, timeslice provides a way
                    to subset results. The subset requested depends on the
                    data type of timeslice:
                        - int: return the a part of the file if it was broken
                               into 2GB chunks (0..N-1)
                        - slice: return the times that correspond to that slice
                                 (i.e., range(ntimes)[timeslice])
                        - list/tuple/set: return specified times where each
                                          time is in the set (0..N-1)
         noscale: Do not apply scaling factors
         vertgrid: vertical coordinate system (options: 'GEOS-5-REDUCED',
                   'GEOS-5-NATIVE', 'MERRA-REDUCED', 'MERRA-NATIVE',
                   'GEOS-4-REDUCED', 'GEOS-4-NATIVE', default 'GEOS-5-REDUCED')
         nogroup: False - use all groups, True - use no groups, list of groups
                  to ignore
        """
        from ._vertcoord import geos_hyai, geos_hybi
        self._ncattrs = ()
        self.noscale = noscale
        self.nogroup = nogroup
        # Read binary data for general header and first datablock header
        header_block = fromfile(bpch_path, dtype='bool',
                                count=_first_header_size)

        # combine data for convenience
        ght = _general_header_type
        dht = _datablock_header_type
        header = (tuple(header_block[:ght.itemsize].view(ght)[0]) +
                  tuple(header_block[ght.itemsize:].view(dht)[0]))

        # Verify that all Fortran unformatted buffers match
        try:
            assert(header[0] == header[2])
            assert(header[3] == header[5])
        except AssertionError:
            raise ValueError("BPCH Files fails header check")

        # Assign data from header to global attributes
        self.ftype = header[1]
        self.toptitle = header[4]
        self.modelname, self.modelres = header[7:9]
        self.halfpolar, self.center180 = header[9:11]
        dummy, dummy, dummy = header[13:16]
        self.start_tau0, self.start_tau1 = header[16:18]
        dummy, dim = header[18:20]
        dummy, dummy = header[20:22]
        for dk, dv in zip('longitude latitude layer'.split(), dim):
            self.createDimension(dk, dv)
        self.createDimension('nv', 2)
        self.createDimension('tnv', 2)
        tracerinfo = tracerinfo or os.path.join(
            os.path.dirname(bpch_path), 'tracerinfo.dat')
        if (
            not hasattr(tracerinfo, 'seek') or
            not hasattr(tracerinfo, 'readlines')
        ):
            if not os.path.exists(tracerinfo) and tracerinfo != ' ':
                tracerinfo = 'tracerinfo.dat'
            if os.path.exists(tracerinfo):
                if os.path.isdir(tracerinfo):
                    tracerinfo = os.path.join(tracerinfo, 'tracerinfo.dat')

            tracerinfo = open(tracerinfo)
        self._tracerinfofile = tracerinfo
        if hasattr(tracerinfo, 'readlines') and hasattr(tracerinfo, 'seek'):
            tracer_data = dict()
            for l in tracerinfo.readlines():
                if l[0] not in ('#', ' '):
                    tkey = int(l[52:61].strip())
                    tdict = dict()
                    tdict['NAME'] = l[:8].strip()
                    tdict['FULLNAME'] = l[9:39].strip()
                    tdict['MOLWT'] = float(l[39:49])
                    tdict['C'] = int(l[49:52])
                    tdict['TRACER'] = int(l[52:61])
                    tdict['SCALE'] = float(l[61:71])
                    tdict['UNIT'] = l[72:].strip()
                    tracer_data[tkey] = tdict

            tracer_names = dict([(k, v['NAME'])
                                 for k, v in tracer_data.items()])
        else:
            warn('Reading file without tracerinfo.dat means that names ' +
                 'and scaling are unknown')
            tracer_data = OrderedDefaultDict(lambda: dict(
                SCALE=1., C=1., MOLWT=1., UNIT='unknown',
                FULLNAME='unknown', NAME='unknown'))
            tracer_names = defaultdictfromkey(lambda key: key)

        diaginfo = diaginfo or os.path.join(
            os.path.dirname(bpch_path), 'diaginfo.dat')
        if (
            not hasattr(diaginfo, 'readlines') or
            not hasattr(diaginfo, 'seek')
        ):
            if not os.path.exists(diaginfo):
                diaginfo = 'diaginfo.dat'
            if os.path.exists(diaginfo):
                if os.path.isdir(diaginfo):
                    diaginfo = os.path.join(diaginfo, 'diaginfo.dat')
            diaginfo = open(diaginfo, 'rt')

        self._diaginfofile = diaginfo

        if hasattr(diaginfo, 'read') and hasattr(diaginfo, 'seek'):
            diag_data = dict([(l[9:49].strip(),
                               dict(offset=int(l[:8]), desc=l[50:].strip()))
                              for l in diaginfo.read().strip().split('\n')
                              if l[0] != '#'])
        else:
            warn('Reading file without diaginfo.dat loses descriptive ' +
                 'information')
            diag_data = defaultdictfromkey(
                lambda key: dict(offset=0, desc=key))

        if (
            len(tracer_names) == 0 and
            not isinstance(tracer_names, defaultdictfromkey)
        ):
            raise IOError("Error parsing %s for Tracer data" % tracerinfo)
        file_size = os.stat(bpch_path).st_size
        offset = _general_header_type.itemsize
        data_types = []
        first_header = None
        keys = []
        self._groups = OrderedDefaultDict(set)
        while first_header is None or \
                offset < file_size:
            header = memmap(bpch_path, offset=offset, shape=(
                1,), dtype=_datablock_header_type, mode=mode)[0]

            group = header[7].decode().strip()
            tracer_number = header[8]
            unit = header[9].strip()
            if not isinstance(diag_data, defaultdictfromkey):
                goffset = diag_data.get(group, {}).get('offset', 0)
                try:
                    tracername = tracer_names[tracer_number + goffset]
                except Exception:
                    # There are some cases like with the adjoint where the
                    # tracer does not have a tracerinfo.dat line.  In this
                    # case, the name matches the tracer with that number
                    # (no offset).  The scaling, however, is not intended
                    # to be used. The unit for adjoint, for instance, is
                    # unitless.
                    if tracer_number not in tracer_names:
                        tracername = str(tracer_number)
                        tracer_data[tracer_number + goffset] = dict(
                            SCALE=1., C=0, UNIT=unit, MOLWT=1.,
                            FULLNAME='unknown', NAME='unknown')
                    else:
                        tracername = tracer_names[tracer_number]
                        tracer_data[tracer_number + goffset] = dict(
                            SCALE=1., C=tracer_data[tracer_number]['C'],
                            UNIT=unit, MOLWT=1.)

            else:
                warn(('%s is not in diaginfo.dat; ' % group) +
                     'names and scaling cannot be resolved')
                goffset = 0
                tracername = str(tracer_number)
                diag_data[group] = dict(offset=0, desc=group)
                tracer_data[tracer_number +
                            goffset] = dict(SCALE=1., C=1., UNIT=unit)

            self._groups[group].add(tracername)
            offset += _datablock_header_type.itemsize + header[-2]

            if first_header is None:
                first_header = header
            elif (
                (header[7], header[8]) == (first_header[7], first_header[8]) or
                offset == file_size
            ):
                if offset == file_size:
                    dim = header[13][::-1]
                    # start = header[14][::-1]
                    data_type = dtype('>i4, %s>f4, >i4' % str(tuple(dim[:])))
                    assert(data_type.itemsize == header[-2])
                    data_types.append(data_type)
                    if self.nogroup is True:
                        keys.append('%s' % (tracername,))
                    elif self.nogroup is False:
                        keys.append('%s_%s' % (group, tracername))
                    elif group in self.nogroup:
                        keys.append('%s' % (tracername,))
                    else:
                        keys.append('%s_%s' % (group, tracername))

                break
            dim = header[13][::-1]
            # start = header[14][::-1]
            data_type = dtype('>i4, %s>f4, >i4' % str(tuple(dim[:])))
            assert(data_type.itemsize == header[-2])
            data_types.append(data_type)
            if self.nogroup is True:
                keys.append('%s' % (tracername,))
            elif self.nogroup is False:
                keys.append('%s_%s' % (group, tracername))
            elif group in self.nogroup:
                keys.append('%s' % (tracername,))
            else:
                keys.append('%s_%s' % (group, tracername))

        # Repeating tracer indicates end of timestep
        keys = [str(k) for k in keys]
        time_type = dtype(
            [(k, dtype([(str('header'), _datablock_header_type),
                        (str('data'), d)]))
             for k, d in zip(keys, data_types)])
        field_shapes = set([v[0].fields['data'][0].fields['f1']
                            [0].shape for k, v in time_type.fields.items()])
        field_levs = set([s_[0] for s_ in field_shapes])
        # field_rows = set([s_[1] for s_ in field_shapes])
        # field_cols = set([s_[2] for s_ in field_shapes])
        field_levs = list(field_levs)
        field_levs.sort()
        for fl in field_levs:
            self.createDimension('layer%d' % fl, fl)

        itemcount = int((float(os.path.getsize(bpch_path)) -
                         _general_header_type.itemsize) // time_type.itemsize)
        if (itemcount % 1) != 0:
            warn("Cannot read whole file; assuming partial time block is at " +
                 "the end; skipping partial time record")
            itemcount = int(np.floor(itemcount))

        # load all data blocks
        try:
            datamap = memmap(bpch_path, dtype=time_type,
                             offset=_general_header_type.itemsize, mode=mode,
                             shape=(itemcount,))
            if timeslice is not None:
                datamap = datamap[timeslice]
        except OverflowError:
            hdrsize = _general_header_type.itemsize
            items = (2 * 1024**3 - hdrsize) // time_type.itemsize
            if timeslice != slice(None):
                filesize = os.stat(bpch_path).st_size
                datasize = (filesize - hdrsize)
                all_times = range(datasize / time_type.itemsize)
                if isinstance(timeslice, int):
                    timeslice = slice(items * timeslice,
                                      items * (timeslice + 1))
                if isinstance(timeslice, (list, tuple, set)):
                    times = timeslice
                else:
                    times = all_times[timeslice]

                outpath = bpch_path + '.tmp.part'
                mint = times[0]
                maxt = times[-1]
                nt = maxt - mint + 1

                if nt > items:
                    warn(('Requested %d items; only returning %d items due ' +
                          'to 2GB limitation') % (nt, items))
                    times = times[:items]

                outfile = open(outpath, 'w')
                infile = open(bpch_path, 'r')
                hdr = infile.read(hdrsize)
                outfile.write(hdr)
                for t in all_times:
                    if t in times:
                        outfile.write(infile.read(time_type.itemsize))
                    else:
                        infile.seek(time_type.itemsize, 1)
                outfile.close()
                datamap = memmap(outpath, dtype=time_type,
                                 mode=mode, offset=hdrsize)
            else:
                datamap = memmap(bpch_path, dtype=time_type, shape=(
                    items,), offset=_general_header_type.itemsize, mode=mode)
                warn('Returning only the first 2GB of data')
        for k in datamap.dtype.names:
            gn = datamap[k]['header']['f7']
            tid = datamap[k]['header']['f8']
            assert((tid[0] == tid).all())
            assert((gn[0] == gn).all())

        layerns = set([datamap[0][k]['header']['f13'][-1]
                       for k in datamap.dtype.names])
        maxlayerns = max(layerns)
        if vertgrid is None:
            if maxlayerns > 48:
                self.vertgrid = vertgrid = 'GEOS-5-NATIVE'
            else:
                self.vertgrid = vertgrid = 'GEOS-5-REDUCED'

            warn('Vertical grid was diagnosted as %s.' % vertgrid)

        # Create variables and dimensions
        self.vertgrid = vertgrid
        # Ap [hPa]
        self.Ap = geos_hyai[self.vertgrid]

        # Bp [unitless]
        self.Bp = geos_hybi[self.vertgrid]

        if max(layerns) > self.Ap.size:
            warn("vertgrid selected (%s) and output layers are not " +
                 "consistent; update to GEOS-5-NATIVE (e.g., bpch(..., " +
                 "vertgrid='GEOS-5-NATIVE') -f \"bpch," +
                 "vertgrid='GEOS-5-NATIVE'\"" % vertgrid)

        layerkeys = ['layer_bounds'] + ['layer%d' % l for l in layerns]
        keys.extend(layerkeys)
        keys.extend(['hyai', 'hyam', 'hybi', 'hybm',
                     'etai_pressure', 'etam_pressure'])

        self.createDimension('layer', self.Ap.size - 1)
        self.createDimension('layer_bounds', self.Ap.size)

        self.variables = _tracer_lookup(parent=self, datamap=datamap,
                                        tracerinfo=tracer_data,
                                        diaginfo=diag_data, keys=keys,
                                        noscale=self.noscale,
                                        nogroup=self.nogroup)
        if self.noscale:
            for vk, v in self.variables.items():
                print(vk, getattr(v, 'scale', 1.))
                if getattr(v, 'scale', 1.) != 1.:
                    warn("Not scaling variables; good for direct writing")

        del datamap
        tdim = self.createDimension('time', self.variables['tau0'].shape[0])
        tdim.setunlimited(True)
        self.groups = dict([(k, _diag_group(self, k, v))
                            for k, v in self._groups.items()])
        for grp in self.groups.values():
            for dk, d in self.dimensions.items():
                dmn = grp.createDimension(dk, len(d))
                dmn.setunlimited(d.isunlimited())
        self.Conventions = "CF-1.6"

    def __repr__(self):
        return PseudoNetCDFFile.__repr__(self) + str(self.variables)


def ncf2bpch(ncffile, outpath, verbose=0):
    outfile = open(outpath, 'wb')
    _general_header_type = np.dtype(dict(
        names=['SPAD1', 'ftype', 'EPAD1', 'SPAD2', 'toptitle', 'EPAD2'],
        formats='>i4, S40, >i4, >i4, S80, >i4'.split()))
    _datablock_header_type = np.dtype(
        dict(names=['SPAD1', 'modelname', 'modelres', 'halfpolar',
                    'center180', 'EPAD1', 'SPAD2', 'category', 'tracerid',
                    'unit', 'tau0', 'tau1', 'reserved', 'dim', 'skip',
                    'EPAD2'],
             formats=['>i4', 'S20', '2>f4', '>i4', '>i4', '>i4', '>i4',
                      'S40', '>i4', 'S40', '>f8', '>f8', 'S40', '6>i4',
                      '>i4', '>i4']))
    varkeys = [k for k, v in ncffile.variables.items()
               if (hasattr(v, 'tracerid') and
                   (k not in metakeys and
                   not k.startswith('layer')))]

    var_types = []
    for varkey in varkeys:
        var = ncffile.variables[varkey]
        data_type = '%s>f' % str(tuple(var[0].shape))
        var_types.append(np.dtype(
            dict(names=['header', 'SPAD1', 'data', 'EPAD1'],
                 formats=[_datablock_header_type, '>i', data_type, '>i'])))
    general_header = np.zeros((1,), dtype=_general_header_type)
    general_header['SPAD1'] = general_header['EPAD1'] = 40
    general_header['SPAD2'] = general_header['EPAD2'] = 80
    general_header['ftype'] = ncffile.ftype.ljust(80)
    general_header['toptitle'] = ncffile.toptitle.ljust(80)
    general_header.tofile(outfile)
    time_data = zeros((1,), dtype=np.dtype(
        dict(names=varkeys, formats=var_types)))

    for varkey in varkeys:
        tdv = time_data[varkey]
        for attrk in _datablock_header_type.names[1:5]:
            tdv['header'][attrk] = getattr(ncffile, attrk)

        tdv['header']['SPAD1'] = tdv['header']['EPAD1'] = 36
        tdv['header']['SPAD2'] = tdv['header']['EPAD2'] = 168

    ttz = zip(ncffile.variables['tau0'], ncffile.variables['tau1'])
    for ti, (tau0, tau1) in enumerate(ttz):
        for varkey in varkeys:
            var = ncffile.variables[varkey]
            if not hasattr(var, 'tracerid'):
                continue
            if verbose:
                print(ti, tau0, tau1, varkey)
            vals = var[ti]
            tdv = time_data[varkey]
            header = tdv['header']
            data = tdv['data']
            header['tau0'] = tau0
            header['tau1'] = tau1
            header['reserved'] = getattr(var, 'reserved', ' ').ljust(40)
            header['tracerid'] = var.tracerid
            header['category'] = var.category.ljust(40)
            header['unit'] = var.base_units
            header['dim'] = (list(vals.shape[::-1]) +
                             [getattr(var, k_, 0) + 1
                              for k_ in 'STARTI STARTJ STARTK'.split()])

            tdv['SPAD1'] = tdv['EPAD1'] = np.prod(vals.shape) * 4
            header['skip'] = tdv['SPAD1'] + 8
            if ncffile.noscale:
                data[:] = vals
            else:
                data[:] = vals / var.scale
        time_data.tofile(outfile)

    outfile.flush()
    outdir = os.path.dirname(outpath)
    tracerpath = os.path.join(outdir, 'tracerinfo.dat')
    diagpath = os.path.join(outdir, 'diaginfo.dat')
    if hasattr(ncffile, '_tracerinfofile'):
        if not os.path.exists(tracerpath):
            ncffile._tracerinfofile.seek(0, 0)
            ncffile._tracerinfofile.seek(0, 0)
            outtrace = open(tracerpath, 'w')
            outtrace.write(ncffile._tracerinfofile.read())
            outtrace.flush()
    if hasattr(ncffile, '_diaginfofile'):
        if not os.path.exists(diagpath):
            ncffile._diaginfofile.seek(0, 0)
            outdiag = open(diagpath, 'w')
            outdiag.write(ncffile._diaginfofile.read())
            outdiag.flush()
    return outfile


registerwriter('bpch', ncf2bpch)


class TestMemmaps(unittest.TestCase):
    def setUp(self):
        from PseudoNetCDF.testcase import geoschemfiles_paths
        self.bpchpath = geoschemfiles_paths['bpch']

    def testBPCH12NCF(self):
        bpchfile = bpch1(self.bpchpath)
        outpath = self.bpchpath + '.nc'
        ncfile = bpchfile.save(outpath)
        for k, ncv in ncfile.variables.items():
            bpv = bpchfile.variables[k]
            np.testing.assert_allclose(ncv[...], bpv[...])
        os.remove(outpath)

    def testNCF2BPCH1(self):
        import warnings
        with warnings.catch_warnings():
            warnings.filterwarnings(
                'ignore', 'Not scaling variables; good for direct writing'
            )
            bpchfile = bpch1(self.bpchpath, noscale=True)
        outpath = self.bpchpath + '.check'
        from PseudoNetCDF.pncgen import pncgen
        pncgen(bpchfile, outpath, inmode='r',
               outmode='w', format='bpch', verbose=0)
        orig = open(self.bpchpath, 'rb').read()
        new = open(outpath, 'rb').read()
        assert(orig == new)
        os.remove(outpath)
        from PseudoNetCDF.sci_var import reduce_dim, slice_dim
        ALD2 = bpchfile.variables['IJ-AVG-$_ALD2']
        ALD2_check = np.array(
            [1.60520077e-02, 1.82803553e-02, 2.00258084e-02, 2.01461259e-02,
             1.84865110e-02, 2.49667447e-02, 2.73083989e-02, 2.87465211e-02,
             2.89694592e-02, 2.87686456e-02, 2.87277419e-02, 3.08121163e-02,
             3.22086290e-02, 3.35262120e-02, 3.41329686e-02, 3.05218045e-02,
             3.30278911e-02, 3.58164124e-02, 3.93186994e-02, 4.15412188e-02,
             1.60520077e-02, 1.82803553e-02, 2.00258084e-02, 2.01461259e-02,
             1.84865110e-02, 2.49667447e-02, 2.73083989e-02, 2.87465211e-02,
             2.89694592e-02, 2.87686456e-02, 2.87277419e-02, 3.08121163e-02,
             3.22086290e-02, 3.35262120e-02, 3.41329686e-02, 3.05218045e-02,
             3.30278911e-02, 3.58164124e-02, 3.93186994e-02, 4.15412188e-02,
             1.60520077e-02, 1.82803553e-02, 2.00258084e-02, 2.01461259e-02,
             1.84865110e-02, 2.49667447e-02, 2.73083989e-02, 2.87465211e-02,
             2.89694592e-02, 2.87686456e-02, 2.87277419e-02, 3.08121163e-02,
             3.22086290e-02, 3.35262120e-02, 3.41329686e-02, 3.05218045e-02,
             3.30278911e-02, 3.58164124e-02, 3.93186994e-02, 4.15412188e-02])\
            .reshape(ALD2.shape)
        slided_reduced_bpchfile = slice_dim(
            reduce_dim(bpchfile, 'layer,mean'), 'time,0')
        pncgen(slided_reduced_bpchfile, outpath, inmode='r',
               outmode='w', format='bpch', verbose=0)
        ALD2_check_slided_reduced = ALD2_check[0].mean(0)[None, None]
        ALD2 = slided_reduced_bpchfile.variables['IJ-AVG-$_ALD2']
        np.testing.assert_allclose(ALD2, ALD2_check_slided_reduced * 1e-9)
        with warnings.catch_warnings():
            warnings.filterwarnings(
                'ignore', 'Not scaling variables; good for direct writing'
            )
            bpchfile = bpch1(outpath)
        ALD2 = bpchfile.variables['IJ-AVG-$_ALD2']
        np.testing.assert_allclose(ALD2, ALD2_check_slided_reduced)

    def testBPCH1(self):
        import warnings
        with warnings.catch_warnings():
            warnings.filterwarnings(
                'ignore', 'Not scaling variables; good for direct writing'
            )
            bpchfile = bpch1(self.bpchpath)
        ALD2 = bpchfile.variables['IJ-AVG-$_ALD2']
        ALD2g = bpchfile.groups['IJ-AVG-$'].variables['ALD2']
        ALD2_check = np.array(
            [1.60520077e-02, 1.82803553e-02, 2.00258084e-02, 2.01461259e-02,
             1.84865110e-02, 2.49667447e-02, 2.73083989e-02, 2.87465211e-02,
             2.89694592e-02, 2.87686456e-02, 2.87277419e-02, 3.08121163e-02,
             3.22086290e-02, 3.35262120e-02, 3.41329686e-02, 3.05218045e-02,
             3.30278911e-02, 3.58164124e-02, 3.93186994e-02, 4.15412188e-02,
             1.60520077e-02, 1.82803553e-02, 2.00258084e-02, 2.01461259e-02,
             1.84865110e-02, 2.49667447e-02, 2.73083989e-02, 2.87465211e-02,
             2.89694592e-02, 2.87686456e-02, 2.87277419e-02, 3.08121163e-02,
             3.22086290e-02, 3.35262120e-02, 3.41329686e-02, 3.05218045e-02,
             3.30278911e-02, 3.58164124e-02, 3.93186994e-02, 4.15412188e-02,
             1.60520077e-02, 1.82803553e-02, 2.00258084e-02, 2.01461259e-02,
             1.84865110e-02, 2.49667447e-02, 2.73083989e-02, 2.87465211e-02,
             2.89694592e-02, 2.87686456e-02, 2.87277419e-02, 3.08121163e-02,
             3.22086290e-02, 3.35262120e-02, 3.41329686e-02, 3.05218045e-02,
             3.30278911e-02, 3.58164124e-02, 3.93186994e-02, 4.15412188e-02])\
            .reshape(ALD2.shape)
        np.testing.assert_allclose(ALD2, ALD2_check)
        np.testing.assert_allclose(ALD2g, ALD2_check)
        np.testing.assert_allclose(bpchfile.variables['hyai'], np.array(
            [0.0, 0.04804826, 6.593752, 13.1348, 19.61311, 26.09201, 32.57081,
             38.98201, 45.33901, 51.69611, 58.05321, 64.36264, 70.62198,
             78.83422, 89.09992, 99.36521, 109.1817, 118.9586, 128.6959,
             142.91, 156.26, 169.609, 181.619, 193.097, 203.259, 212.15,
             218.776, 223.898, 224.363, 216.865, 201.192, 176.93, 150.393,
             127.837, 108.663, 92.36572, 78.51231, 56.38791, 40.17541,
             28.36781, 19.7916, 9.292942, 4.076571, 1.65079, 0.6167791,
             0.211349, 0.06600001, 0.01], 'f'))
        np.testing.assert_allclose(bpchfile.variables['hybi'], np.array(
            [1.0, 0.984952, 0.963406, 0.941865, 0.920387, 0.898908, 0.877429,
             0.856018, 0.8346609, 0.8133039, 0.7919469, 0.7706375, 0.7493782,
             0.721166, 0.6858999, 0.6506349, 0.6158184, 0.5810415, 0.5463042,
             0.4945902, 0.4437402, 0.3928911, 0.3433811, 0.2944031, 0.2467411,
             0.2003501, 0.1562241, 0.1136021, 0.06372006, 0.02801004,
             0.006960025, 8.175413e-09, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], 'f'))
        etai_check = np.array(
            [1.01325000e+03, 9.98050662e+02, 9.82764882e+02, 9.67479511e+02,
             9.52195238e+02, 9.36910541e+02, 9.21625744e+02, 9.06342248e+02,
             8.91059167e+02, 8.75776287e+02, 8.60493406e+02, 8.45211087e+02,
             8.29929441e+02, 8.09555669e+02, 7.84087994e+02, 7.58621022e+02,
             7.33159694e+02, 7.07698900e+02, 6.82238631e+02, 6.44053520e+02,
             6.05879758e+02, 5.67705907e+02, 5.29549900e+02, 4.91400941e+02,
             4.53269420e+02, 4.15154739e+02, 3.77070069e+02, 3.39005328e+02,
             2.88927351e+02, 2.45246173e+02, 2.08244245e+02, 1.76930008e+02,
             1.50393000e+02, 1.27837000e+02, 1.08663000e+02, 9.23657200e+01,
             7.85123100e+01, 5.63879100e+01, 4.01754100e+01, 2.83678100e+01,
             1.97916000e+01, 9.29294200e+00, 4.07657100e+00, 1.65079000e+00,
             6.16779100e-01, 2.11349000e-01, 6.60000100e-02, 1.00000000e-02])
        np.testing.assert_allclose(bpchfile.variables['etai_pressure'],
                                   etai_check)

    def runTest(self):
        pass


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
import os
import sys
import numpy as np
import datetime
from argparse import ArgumentParser
from urllib.parse import urlencode
from urllib.request import urlopen, Request
from io import BytesIO
import gzip
import tempfile
import getpass
from netCDF4 import MFDataset, Dataset
from shapely.wkt import loads
from shapely.geometry import Point, Polygon
from shapely.prepared import prep

from PseudoNetCDF.pncgen import pncgen
from PseudoNetCDF.sci_var import Pseudo2NetCDF, PseudoNetCDFFile

parser = ArgumentParser()
parser.add_argument('-v', '--verbose', default=0,
                    action='count', help='Show progress')
parser.add_argument('--checkflags', type=str, default='CSVKQkG',
                    help='MADIS DD flags (default passes at least stage 1 or in Accept list)')
parser.add_argument('--dtmin', dest='dt_min', default=-1800.,
                    help='Seconds from nominal time negative')
parser.add_argument('--dtmax', dest='dt_max', default=1800.,
                    help='Seconds from nominal time positive')
parser.add_argument('--no-humditity', dest='humidity', default=True,
                    action='store_false', help='Disable calculation of specific humidity')
parser.add_argument('--username', default=None, help='MADIS username')
parser.add_argument('--password', default=None, help='MADIS password')
parser.add_argument('--cache', default=False,
                    action='store_true', help='cache MADIS files')
parser.add_argument('--type', choices=["sao", "metar", "maritime", "mesonet", "raob",
                                       "acarsProfiles", "profiler", "profiler", "hydro"], help='MADIS data types')
spacegroup = parser.add_mutually_exclusive_group(required=False)
spacegroup.add_argument('--gridcro2d', dest='GRIDCRO2D',
                        help='CMAQ MCIP GRIDCRO2D file or any file that has LAT and LON variables')
spacegroup.add_argument('--wktpolygon')
parser.add_argument('--START_DATE', type=lambda x: datetime.datetime.strptime(x,
                                                                              '%Y-%m-%d %H:%M'), help='YYYY-MM-DD HH:MM')
parser.add_argument('--END_DATE', type=lambda x: datetime.datetime.strptime(x,
                                                                            '%Y-%m-%d %H:%M'), help='YYYY-MM-DD HH:MM')
parser.add_argument('--TSTEP', default='1 hours')
parser.add_argument('--hourlypath', help='Path for optional hourly output.')
parser.add_argument('rawpath', help='Outpath for timeObs output')

args = parser.parse_args()

if args.GRIDCRO2D is not None:
    gridcrof = Dataset(args.GRIDCRO2D)
    lon = gridcrof.variables['LON'][0, 0]
    lat = gridcrof.variables['LAT'][0, 0]
    llcrnrlon = lon[:, 0].max()
    urcrnrlon = lon[:, -1].min()
    llcrnrlat = lat[0, :].max()
    urcrnrlat = lat[-1, :].min()
    bounds = Polygon([(llcrnrlon, llcrnrlat), (llcrnrlon, urcrnrlat),
                      (urcrnrlon, urcrnrlat), (urcrnrlon, llcrnrlat)])
    prep_bounds = prep(bounds)
elif args.wktpolygon is not None:
    bounds = loads(args.wktpolygon)
    prep_bounds = prep(bounds)
else:
    bounds = None
    prep_bounds = None

level1 = dict(zip(("sao", "metar", "maritime", "mesonet", "raob", "acarsProfiles", "profiler", "profiler",
                   "hydro"), ("point", "point", "point", "LDAD", "point", "point", "point", "LDAD", "LDAD")))[args.type]
level3 = dict(zip(("sao", "metar", "maritime", "mesonet", "raob", "acarsProfiles", "profiler", "profiler", "hydro"),
                  ("netcdf", "netcdf", "netcdf", "netCDF", "netcdf", "netcdf", "netcdf", "netCDF", "netCDF")))[args.type]

authvalues = dict(username=args.username, password=args.password)
auth = urlencode(authvalues)

template = 'ftp://pftp.madis-data.noaa.gov/archive/%%Y/%%m/%%d/%(level1)s/%(level2)s/%(level3)s/%%Y%%m%%d_%%H%%M.gz' % dict(
    level2=args.type, level3=level3, level1=level1)

times = [args.START_DATE]
now = times[0]
inc = datetime.timedelta(hours=1)
while now < args.END_DATE:
    now = now + inc
    times.append(now)


timefiles = {}
infiles = []
for time in times:
    if args.verbose > 0:
        print(time, 'start')
    url = time.strftime(template)
    cachedpath = time.strftime(os.path.join(
        'madis', 'archive', '%Y', '%m', '%d', level1, args.type, level3, os.path.basename(url)))
    if not args.cache or not os.path.exists(cachedpath):
        if args.username is None:
            args.username = input('MADIS username:')

        if args.password is None:
            args.password = getpass.getpass(
                prompt='MADIS password: ', stream=None)
        if args.verbose > 0:
            print('downloading: ' + url)
        req = Request(url, data=auth)
        try:
            ftpf = urlopen(req)
        except Exception:
            print('Failed to retrieve ' + url +
                  '; continuing', file=sys.stderr)
            continue
        # create BytesIO temporary file
        if args.cache:
            dirpath = os.path.dirname(cachedpath)
            if not os.path.exists(dirpath):
                os.makedirs(dirpath)
            compressedFile = open(cachedpath, 'w+b')
        else:
            compressedFile = BytesIO()
        compressedFile.write(ftpf.read())
        compressedFile.flush()
    else:
        if args.verbose > 0:
            print('using cache ' + cachedpath)
        compressedFile = open(cachedpath, 'r+b')
    compressedFile.seek(0)
    decompressedFile = gzip.GzipFile(fileobj=compressedFile, mode='rb')
    diskf = tempfile.NamedTemporaryFile('w+b')
    infiles.append(diskf)
    diskf.write(decompressedFile.read())
    diskf.flush()
    if args.verbose > 0:
        print(time, 'end')


ncff = MFDataset([inf.name for inf in infiles], 'r')
lat = ncff.variables['latitude'][:]
lon = ncff.variables['longitude'][:]
points = zip(lon, lat)
if prep_bounds is None:
    found_point_ids = range(len(points))
else:
    found_point_ids = []
    for pi, point in enumerate(points):
        isin = prep_bounds.contains(Point(*point))
        if isin:
            found_point_ids.append(pi)
            if args.verbose > 1:
                print(point, isin)
        elif args.verbose > 2:
            print(point, isin)

varkeys = ['temperature', 'windDir', 'windSpeed', 'dewpoint', 'altimeter']
vardds = [k + 'DD' for k in varkeys]

if args.verbose > 1:
    print('Subset variables')

getvarkeys = varkeys + vardds + \
    ['stationName', 'timeObs', 'timeNominal', 'elevation', 'latitude', 'longitude']

if args.verbose > 1:
    print('Slicing files')

p2p = Pseudo2NetCDF(verbose=0)
outfile = PseudoNetCDFFile()
p2p.addDimensions(ncff, outfile)
outfile.createDimension('recNum', len(found_point_ids))
p2p.addGlobalProperties(ncff, outfile)

for vark in getvarkeys:
    p2p.addVariable(ncff, outfile, vark, data=False)

for vark in getvarkeys:
    invar = ncff.variables[vark]
    outvar = outfile.variables[vark]
    recid = list(invar.dimensions).index('recNum')
    outvar[:] = invar[:].take(found_point_ids, recid)

if args.humidity:
    varkeys.append('specificHumidity')

if args.verbose > 1:
    print('Making output files')

maxlen = len(outfile.dimensions['maxStaNamLen'])
stationNames = outfile.variables['stationName'][:].view(
    'S%d' % maxlen).squeeze()
timeObs = outfile.variables['timeObs']
timeNominal = outfile.variables['timeNominal']

ustations = np.unique(stationNames)
utimes = np.sort(np.unique(timeObs[:]))
unomtimes = np.sort(np.unique(timeNominal[:]))

obstimefile = Dataset(args.rawpath, 'w', format='NETCDF4_CLASSIC')
obstimefile.createDimension('time', utimes.size)
obstimefile.createDimension('timeNominal', unomtimes.size)
obstimefile.createDimension('stations', ustations.size)
obstimefile.createDimension('maxStaNamLen', maxlen)

inelev = outfile.variables['elevation']
inlat = outfile.variables['latitude']
inlon = outfile.variables['longitude']

outstat = obstimefile.createVariable(
    'stationNames', 'S1', ('stations', 'maxStaNamLen'))
outstat[:] = ustations.view('S1')

outlat = obstimefile.createVariable('latitude', 'f', ('stations',))
for pk in inlat.ncattrs():
    setattr(outlat, pk, getattr(inlat, pk))

outlon = obstimefile.createVariable('longitude', 'f', ('stations',))
for pk in inlon.ncattrs():
    setattr(outlon, pk, getattr(inlon, pk))

outelev = obstimefile.createVariable('elevation', 'f', ('stations',))
for pk in inelev.ncattrs():
    setattr(outelev, pk, getattr(inelev, pk))

if args.humidity:
    from PseudoNetCDF.derived.met import wmr_ptd
    outfile.variables['specificHumidity'] = wmr_ptd(
        outfile.variables['altimeter'] / 100., outfile.variables['dewpoint'] - 273.15, gkg=True)
    outfile.variables['specificHumidityDD'] = outfile.variables['dewpointDD']

station_idx = (stationNames[:, None] == ustations[None, :]).argmax(1)
time_idx = (timeObs[:][:, None] == utimes[None, :]).argmax(1)

time = obstimefile.createVariable('time', 'd', ('time',))
for pk in timeObs.ncattrs():
    setattr(time, pk, getattr(timeObs, pk))
time.units = time.units.replace('.0', '')

nomtime = obstimefile.createVariable('timeNominal', 'd', ('timeNominal',))
for pk in timeNominal.ncattrs():
    setattr(nomtime, pk, getattr(timeNominal, pk))
nomtime.units = nomtime.units.replace('.0', '')

for vark in varkeys:
    invar = outfile.variables[vark]
    outvar = obstimefile.createVariable(
        vark, invar.dtype, ('time', 'stations'), fill_value=-999)
    for pk in invar.ncattrs():
        setattr(outvar, pk, getattr(invar, pk))

tmpvals = np.zeros_like(outlat)
tmpvals[station_idx] = inlat[:]
outlat[:] = tmpvals[:]
tmpvals = np.zeros_like(outlon)
tmpvals[station_idx] = inlon[:]
outlon[:] = tmpvals[:]
tmpvals = np.zeros_like(outelev)
tmpvals[station_idx] = inelev[:]
outelev[:] = tmpvals[:]
time[:] = utimes[:]
nomtime[:] = unomtimes

for vark in varkeys:
    invar = outfile.variables[vark]
    outvar = obstimefile.variables[vark]
    indd = np.char.decode(outfile.variables[vark + 'DD'][:], 'ascii')
    inval = invar[:]
    tmpvals = np.zeros_like(outvar[:])
    passes = np.sum([indd[:] == flag for flag in args.checkflags], axis=0) > 0
    fails = ~passes
    tmpvals[time_idx, station_idx] = np.ma.masked_where(fails, invar[:])
    outvar[:] = tmpvals

if args.verbose > 1:
    print('Saving hourly data')


if args.hourlypath is not None:
    dt_min = args.dt_min
    dt_max = args.dt_max

    from PseudoNetCDF.sci_var import Pseudo2NetCDF, PseudoNetCDFFile
    p2p = Pseudo2NetCDF(verbose=False)
    nomtimefile = PseudoNetCDFFile()
    p2p.addDimensions(obstimefile, nomtimefile)
    nomtimefile.createDimension('time', unomtimes.size)
    p2p.addGlobalProperties(obstimefile, nomtimefile)

    for vark in varkeys:
        p2p.addVariable(obstimefile, nomtimefile, vark, data=False)
        outvar = nomtimefile.variables[vark]
        var = obstimefile.variables[vark]
        timem = (var[:] / var[:]) * time[:][:, None]
        dt = np.ma.MaskedArray(unomtimes[:, None, None]) - timem[None, :, :]
        nomtime_idx, points_idx = np.indices(dt[:, 0].shape)
        time_idx = np.abs(dt).argmin(1)
        outvals = var[:][time_idx, points_idx]
        min_dt = dt[nomtime_idx, time_idx, points_idx]
        dt_mask = (min_dt < dt_min) | (min_dt > dt_max)
        outvals = np.ma.masked_where(dt_mask, outvals)
        outvar[:] = outvals[:]

    pncgen(nomtimefile, args.hourlypath, verbose=0)
# :DD_long_name = "QC data descriptor model:  QC summary values" ;
# :DD_reference = "AWIPS Technique Specification Package (TSP) 88-21-R2" ;
# :DD_values = "Z,C,S,V,X,Q,K,k,G, or B" ;
# :DD_value_Z = "No QC applied" ;
# :DD_value_C = "Passed QC stage 1" ;
# :DD_value_S = "Passed QC stages 1 and 2" ;
# :DD_value_V = "Passed QC stages 1, 2 and 3" ;
# :DD_value_X = "Failed QC stage 1" ;
# :DD_value_Q = "Passed QC stage 1, but failed stages 2 or 3 " ;
# :DD_value_K = "Passed QC stages 1, 2, 3, and 4" ;
# :DD_value_k = "Passed QC stage 1,2, and 3, failed stage 4 " ;
# :DD_value_G = "Included in accept list" ;
# :DD_value_B = "Included in reject list" ;



from ._files import PseudoNetCDFFile
from collections import OrderedDict


class WrapDict(OrderedDict):
    def __init__(self, other={}):
        self._other = other
        self._mine = OrderedDict()

    def __getitem__(self, k):
        if k in self._mine:
            return self._mine[k]
        elif k in self._other:
            return self._other[k]
        else:
            raise KeyError(
                '{0} not in dictionary or wrapped dictionary'.format(k))

    def items(self):
        for k, v in self._other.items():
            if k not in self._mine:
                yield k, v

        for k, v in self._mine.items():
            yield k, v

    def keys(self):
        for k in self._other.keys():
            if k not in self._mine:
                yield k

        for k in self._mine.keys():
            yield k

    def __len__(self):
        return len(self._mine) + len(self._other)

    def __iter__(self):
        return self.keys()

    def __contains__(self, k):
        for myk in self.keys():
            if k == myk:
                return True
        else:
            return False

    def __setitem__(self, k, v):
        self._mine[k] = v

    def __delitem__(self, k):
        if k in self._mine:
            del self._mine[k]

        if k in self._other:
            del self._other[k]

    def __repr__(self):
        outf = OrderedDict()
        for key in self.keys():
            outf[key] = self[key]
        return outf.__repr__()

    def __str__(self):
        outf = OrderedDict()
        for key in self.keys():
            outf[key] = self[key]
        return outf.__str__()


class WrapPNC(PseudoNetCDFFile):
    @classmethod
    def isMine(cls, path, *args, **kwds):
        return False

    def __init__(self, *args, **kwds):
        # from PseudoNetCDF import pncopen
        self._file = args[0]  # pncopen(*args, **kwds)
        for k in self._file.ncattrs():
            setattr(self, k, getattr(self._file, k))
        self.dimensions = WrapDict(self._file.dimensions)
        self.variables = WrapDict(self._file.variables)


if __name__ == '__main__':
    cpath = 'CONC/CCTM_CONC_v52_hdifupdate_intel17.0_HEMIS_cb6_20160101'
    wfile = WrapPNC(cpath)



__all__ = ['write_wind']
__doc__ = """
.. _Write
:mod:`Write` -- CAMx wind writer
============================================

.. module:: Write
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` writer for CAMx
              wind files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

import numpy as np

# This Package modules
from PseudoNetCDF.camxfiles.timetuple import timeadd, timerange
from PseudoNetCDF.camxfiles.FortranFileUtil import writeline
from PseudoNetCDF._getwriter import registerwriter

"""
hour,idate,lstagger
Loop from 1 to nlay layers:
    ((uwind(i,j,k),i=1,nx),j=1,ny)
    ((vwind(i,j,k),i=1,nx),j=1,ny)
((dummy,i=1,nx),j=1,ny)
"""


def ncf2wind(ncffile, outpath, tflag='TFLAG'):
    outfile = open(outpath, 'wb')
    varkeys = [vk for vk in ['U', 'V'] if vk in ncffile.variables.keys()]
    nzcl = len(ncffile.dimensions['LAY'])
    for di, (d, t) in enumerate(ncffile.variables[tflag][:, 0, :]):
        t = np.array(t // 100, ndmin=1).astype('>f')
        d = np.array(d, ndmin=1).astype('>i')
        d = (d % (d // 100000 * 100000)).astype('>i')
        lstag = ncffile.LSTAGGER
        buf = np.array([12], dtype='>i').tostring()
        outfile.write(buf + t.tostring() + d.tostring() +
                      lstag.tostring() + buf)
        for zi in range(nzcl):
            for varkey in varkeys:
                vals = ncffile.variables[varkey][di, zi].astype('>f')
                buf = np.array([(vals.size) * 4],
                               ndmin=1).astype('>i').tostring()
                outfile.write(buf)
                vals.tofile(outfile)
                outfile.write(buf)
        vals = np.array(0, dtype='>i')
        buf = np.array([(vals.size) * 4], ndmin=1).astype('>i').tostring()
        outfile.write(buf)
        vals.astype('>f').tofile(outfile)
        outfile.write(buf)
    outfile.flush()
    return outfile


registerwriter('camxfiles.wind', ncf2wind)
registerwriter('wind', ncf2wind)


def write_wind(sdate, stime, time_step, vals, lstagger=None):
    """
    Takes an iterable and some defining information
    and creates a CAMx read wind file

    sdate - integer start date
    stime - float start time
    time_step - float increment between times

    vals - array axes time,uv,xy,z
    """
    wind_string = ""
    edate, etime = timeadd((sdate, stime), (0, vals.shape[0] * time_step))

    for i, (d, t) in enumerate(timerange((sdate, stime), (edate, etime),
                                         time_step)):
        if lstagger is not None:
            wind_string += writeline((t, d, lstagger), "fii")
        else:
            wind_string += writeline((t, d), "fi")

        for k in range(vals.shape[-1]):
            for uv in range(2):
                wind_string += writeline(vals[i, uv, ..., k],
                                         "f" * vals.shape[-2])
        wind_string += writeline((0,), "i")
    return wind_string



import unittest
from warnings import warn
from datetime import datetime
import sys
import os

import numpy as np

from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariables
from PseudoNetCDF.sci_var import PseudoNetCDFVariable


_geos_units = dict(ALBEDO='unitless',
                   CLDTOT='unitless',
                   EFLUX='W/m2',
                   EVAP='unitless',
                   FRLAKE='unitless',
                   FRLAND='unitless',
                   FRLANDIC='unitless',
                   FROCEAN='unitless',
                   GRN='unitless',
                   GWETROOT='unitless',
                   GWETTOP='unitless',
                   HFLUX='W/m2',
                   LAI='m2/m2',
                   LWGNET='W/m2',
                   LWTUP='W/m2',
                   PARDF='W/m2',
                   PARDR='W/m2',
                   PBLH='m',
                   PRECANV='kg/m2/s',
                   PRECCON='kg/m2/s',
                   PRECTOT='kg/m2/s',
                   PRECSNO='kg/m2/s',
                   QV2M='kg/kg',
                   SNOMAS='m',
                   SNODP='m',
                   SWGNET='W/m2',
                   T2M='K',
                   TROPP='hPa',
                   TSKIN='K',
                   U10M='m/s',
                   USTAR='m/s',
                   V10M='m/s',
                   Z0M='m',
                   DQRCON='kg/m2/s',
                   DQRLSC='kg/m2/s',
                   DTRAIN='kg/m2/s',
                   QL='unitless',
                   QI='unitless',
                   OPTDEPTH='unitless',
                   TAUCLI='unitless',
                   TAUCLW='unitless',
                   CLOUD='unitless',
                   PV='K*m2/kg/s',
                   OMEGA='Pa/s',
                   QV='kg/kg',
                   RH='unitless',
                   T='K',
                   U='m/s',
                   V='m/s',
                   DQIDTMST='kg/kg/s',
                   DQLDTMST='kg/kg/s',
                   DQVDTMST='kg/kg/s',
                   MOISTQ='kg/kg/s',
                   CMFMC='kg/m2/s',
                   LWI='m/s',
                   PS='hPa',
                   SLP='hPa',
                   TO3='DU',
                   TTO3='DU')


def lump(data, ap, levels):
    ldata = data[:, levels]
    lweights = -np.diff(ap)[levels][None, :, None, None]
    return (ldata * lweights / lweights.sum(1)).sum(1)


class geos(PseudoNetCDFFile):
    def __init__(self, path, mode='r', reduced=True):
        from _vertcoord import geos_hyai, geos_hybi, geos_etam, geos_etai
        infile = open(path, 'r')
        infile.seek(0, 0)
        fsize = os.path.getsize(path)
        lasttype = ''
        datatypes = [('name', '>i4, >S8, >i4')]
        names = []
        first = True
        # values will be found before use
        name = nrow = ncol = nlay = nlay_stag = nlay_in = nlay_in_stag = None
        while infile.tell() < fsize:
            ipad = np.fromfile(infile, dtype='>i', count=1)
            dtype = '>S8' if ipad == 8 else '>f'
            if dtype == '>f':
                elem = ipad / 4
                assert((ipad % 4) == 0)
            else:
                elem = 1
            if first:
                pass
            else:
                if elem == 1:
                    pass  # shape unknown
                elif elem == (nrow * ncol + 2):
                    shape = (nrow, ncol)
                elif elem == (nrow * ncol * nlay_in + 2):
                    shape = (nlay_in, nrow, ncol)
                elif elem == (nrow * ncol * nlay_in_stag + 2):
                    shape = (nlay_in_stag, nrow, ncol)
                else:
                    raise IOError('Expected %d, %d, or %d elements, got %d' % (
                        nrow * ncol + 2, nrow * ncol * nlay + 2,
                        nrow * ncol * nlay_stag + 2, elem))
            if dtype == '>S8':
                data = np.fromfile(infile, dtype=dtype, count=elem)
            else:
                infile.seek(ipad, 1)
            # data = np.fromfile(infile, dtype = dtype, count = elem)
            # date, time = data[:2]
            # if dtype != '>c':
            #    data = data[2:].reshape(shape)
            epad = np.fromfile(infile, dtype='>i', count=1)
            if lasttype == '>S8' and name[:2] not in ('G5', 'G4'):
                datatypes.append(
                    (name,
                     ('>i4, >S8, >i4, >i4, >i4, >i4, %s>f, >i4' %
                      str(tuple(shape)))))
            else:
                name = data[0].strip()
                if first:
                    first = False
                    if name[:2] == 'G5':
                        etak = 'GEOS-5-' + ('REDUCED' if reduced else 'NATIVE')
                        etafk = 'GEOS-5-' + 'NATIVE'
                    elif name[:2] == 'G4':
                        etak = 'GEOS-4-' + ('REDUCED' if reduced else 'NATIVE')
                        etafk = 'GEOS-4-' + 'NATIVE'
                    else:
                        raise ValueError('Unknown type %s' % name)
                    self.gtype = etak
                    # Ap [hPa]
                    self.Ap_REDUCED = geos_hyai[etak]
                    # Bp [unitless]
                    self.Bp_REDUCED = geos_hybi[etak]
                    # Ap full [hPa]
                    self.Ap_NATIVE = geos_hyai[etafk]
                    # Bp full [unitless]
                    self.Bp_NATIVE = geos_hybi[etafk]
                    eta_m = geos_etam[etak]
                    eta_i = geos_etai[etak]

                    nlay_in = self.Ap_NATIVE.size - 1
                    nlay_in_stag = self.Ap_NATIVE.size
                    if reduced:
                        self.Ap = self.Ap_REDUCED
                        self.Bp = self.Bp_REDUCED
                    else:
                        self.Ap = self.Ap_NATIVE
                        self.Bp = self.Bp_NATIVE
                    nlay = self.Ap.size - 1
                    nlay_stag = self.Ap.size

                    if name[3:5] == '22':
                        lone1d = np.arange(-181.25, 180, 2.5)
                        lone = lone1d.repeat(2, 0)[1:-1].reshape(-1, 2)
                        longitude_bounds = lone
                        late1d = np.append(np.append(-90.,
                                                     np.arange(-89., 90., 2.)),
                                           90.)
                        late = late1d.repeat(2, 0)[1:-1].reshape(-1, 2)
                        latitude_bounds = late
                        nrow = 91
                        ncol = 144
                    elif name[3:5] == '45':
                        lone1d = np.arange(-182.5, 180, 5)
                        lone = lone1d.repeat(2, 0)[1:-1].reshape(-1, 2)
                        longitude_bounds = lone
                        late1d = np.append(np.append(-90.,
                                                     np.arange(-88., 90., 4.)),
                                           90.)
                        late = late1d.repeat(2, 0)[1:-1].reshape(-1, 2)
                        latitude_bounds = late
                        nrow = 46
                        ncol = 72
                    else:
                        raise ValueError('Unknown type %s' % name)

                    longitude = longitude_bounds.mean(1)
                    latitude = latitude_bounds.mean(1)
                if name in names:
                    break
                names.append(name)
            lasttype = dtype
            assert(ipad == epad)
        if 'a3' in path:
            nsteps = 8
        elif 'a6' in path or 'i6' in path:
            nsteps = 4
        datatype = [datatypes[0], ('data', np.dtype(datatypes[1:]), nsteps)]
        data = np.memmap(path, dtype=np.dtype(datatype), mode=mode)
        d = self.createDimension('time', nsteps)
        d.setunlimited(True)
        self.createDimension('latitude', nrow)
        self.createDimension('longitude', ncol)
        self.createDimension('layer', nlay)
        self.createDimension('layer_stag', nlay_stag)
        self.createDimension('nv', 2)
        self.title = data['name']['f1'][0].strip()

        def getem(key):
            thisblock = data[0]['data'][key]
            thisdata = thisblock['f6']
            assert((thisblock['f3'] == thisblock['f7']).all())
            if len(thisdata.shape) == 3:
                dims = ('time', 'latitude', 'longitude')
            elif thisdata.shape[1] == nlay_in:
                dims = ('time', 'layer', 'latitude', 'longitude')
            elif thisdata.shape[1] == nlay_in_stag:
                dims = ('time', 'layer_stag', 'latitude', 'longitude')
            else:
                raise ValueError('Wrong layers got %d not %d or %d' %
                                 (thisdata.shape[1], nlay, nlay_stag))
            unit = _geos_units.get(key, '')
            if dims != ('time', 'latitude', 'longitude'):
                thisdatain = thisdata
                thisdata = np.zeros([len(self.dimensions[k]) for k in dims],
                                    dtype=thisdata.dtype)
                if reduced:
                    if self.gtype == 'GEOS-4-REDUCED':
                        # !----------------------------------------------------
                        # ! GEOS-4: Lump 55 levels into 30 levels, starting
                        # ! above L=20
                        # ! Lump levels in groups of 2, then 4. (cf. Mat Evans)
                        # !----------------------------------------------------
                        #
                        # ! Lump 2 levels together at a time
                        lump_groups = [[0, ], [1, ], [2, ], [3, ], [4, ],
                                       [5, ], [6, ], [7, ], [8, ], [9, ],
                                       [10, ], [11, ], [12, ], [13, ], [14, ],
                                       [15, ], [16, ], [17, ], [18, ]] + \
                                      [[19, 20], [21, 22],
                                       [23, 24], [25, 26]] + \
                                      [[27, 28, 29, 30], [31, 32, 33, 34],
                                       [35, 36, 37, 38], [39, 40, 41, 42],
                                       [43, 44, 45, 46],
                                       [47, 48, 49, 50],
                                       [51, 52, 53, 54]]

                    elif self.gtype == 'GEOS-5-REDUCED':
                        # !----------------------------------------------------
                        # ! GEOS-5/MERRA: Lump 72 levels into 47 levels,
                        # ! starting above L=36.  Lump levels in groups of 2,
                        # ! then 4. (cf. Bob Yantosca)
                        # !----------------------------------------------------
                        #
                        # ! Lump 2 levels together at a time
                        lump_groups = [[0, ], [1, ], [2, ], [3, ], [4, ],
                                       [5, ], [6, ], [7, ], [8, ], [9, ],
                                       [10, ], [11, ], [12, ], [13, ], [14, ],
                                       [15, ], [16, ], [17, ], [18, ], [19, ],
                                       [20, ], [21, ], [22, ], [23, ], [24, ],
                                       [25, ], [26, ], [27, ], [28, ], [29, ],
                                       [30, ], [31, ], [32, ], [33, ], [34, ],
                                       [35, ]] + \
                                      [[36, 37], [38, 39],
                                       [40, 41], [42, 43]] + \
                                      [[44, 45, 46, 47], [48, 49, 50, 51],
                                       [52, 53, 54, 55], [56, 57, 58, 59],
                                       [60, 61, 62, 63], [64, 65, 66, 67],
                                       [68, 69, 70, 71]]
                    else:
                        raise ValueError('Cannot reduce %s' % self.gtype)
                    assert(len(lump_groups) == nlay)

                    for li, lump_group in enumerate(lump_groups):
                        if (len(lump_group) == 1 or
                                dims[1] == 'layer_stag' or key == 'PLE'):
                            thisdata[:, li] = thisdatain[:, lump_group[0]]
                        elif dims[1] == 'layer':
                            # assumes lumping only happens above pure eta
                            # true for (GEOS4 and GEOS5)
                            thisdata[:, li] = lump(
                                thisdatain, self.Ap_NATIVE, lump_group)
                        else:
                            raise ValueError('huh?')
                    else:
                        if dims[1] == 'layer_stag':
                            thisdata[:, li + 1] = thisdatain[:, lump_group[-1]]

                else:
                    thisdata = thisdatain
            return PseudoNetCDFVariable(self, key, 'f', dims, values=thisdata,
                                        units=unit, long_name=key.ljust(16))
        self.variables = PseudoNetCDFVariables(keys=names[1:], func=getem)
        dates = data[0]['data'][names[1]]['f4']
        times = data[0]['data'][names[1]]['f5']
        years, months, days = dates // 10000, dates % 10000 // 100, dates % 100
        hours, minutes = times // 10000, times % 10000 // 100
        seconds = times % 100
        self.createVariable('layer', 'f', ('layer',),
                            values=eta_m[:], units='hybrid pressure-sigma',
                            bounds='layer_bounds')
        self.createVariable('layer_bounds', 'f', ('layer', 'nv'),
                            values=eta_i[:].repeat(2, 0)[1:-1].reshape(-1, 2),
                            units='hybrid pressure-sigma')

        rdate = datetime(1985, 1, 1, 0, 0, 0)

        def ymdhms2reldate(y, m, d, H, M, S):
            dates = datetime(y, m, d, H, M, S)
            rdates = dates - rdate
            seconds = rdates.total_seconds()
            hours = seconds / 3600.
            return hours

        enumdateparts = zip(years, months, days, hours, minutes, seconds)
        self.createVariable('time', 'f', ('time',),
                            values=np.array([ymdhms2reldate(*ymdHMS)
                                             for ymdHMS in enumdateparts]),
                            units='hours since 1985-01-01 00:00:00 UTC',
                            base_units='hours since 1985-01-01 00:00:00 UTC',
                            long_name='time', standard_name='time')
        self.createVariable('longitude', 'f', ('longitude',), values=longitude,
                            units='degrees_east', standard_name='longitude',
                            bounds='longitude_bounds')
        self.createVariable('longitude_bounds', 'f', ('longitude', 'nv'),
                            values=longitude_bounds, units='degrees_east',
                            standard_name='longitude')
        self.createVariable('latitude', 'f', ('latitude',), values=latitude,
                            units='degrees_north', standard_name='latitude',
                            bounds='latitude_bounds')
        self.createVariable('latitude_bounds', 'f', ('latitude', 'nv'),
                            values=latitude_bounds, units='degrees_north',
                            standard_name='latitude')


class TestMemmaps(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testGEOS(self):
        warn('Test not implemented')


if __name__ == '__main__':
    path = sys.argv[1]
    f = geos(path)



#!/usr/bin/env python
from __future__ import print_function
# Run this script with pnc options
import os

from PseudoNetCDF.pncload import PNCConsole
from PseudoNetCDF.coordutil import getmap, getlatbnds, getlonbnds
from PseudoNetCDF.coordutil import getybnds, getxbnds
from PseudoNetCDF.sci_var import getvarpnc, slice_dim

import warnings
import numpy as np
import matplotlib.pyplot as plt
warn = warnings.warn
pl = plt
LogFormatter = plt.matplotlib.ticker.LogFormatter
BoundaryNorm = plt.matplotlib.colors.BoundaryNorm
# from PseudoNetCDF.plotutil import *


def makemaps(args):
    ifiles = args.ifiles
    cbar = None
    ifile = ifiles[0]
    if args.iter != []:
        ifile, = ifiles
        ifiles = []
        for dimk in args.iter:
            ifiles += [slice_dim(getvarpnc(ifile, None), '%s,%d' % (dimk, i))
                       for i in range(len(ifile.dimensions[dimk]))]
    ax = plt.gca()
    map = getmap(ifile, resolution=args.resolution)
    if args.coastlines:
        map.drawcoastlines(ax=ax)
    if args.countries:
        map.drawcountries(ax=ax)
    if args.states:
        map.drawstates(ax=ax)
    if args.counties:
        map.drawcounties(ax=ax)
    for si, shapefile in enumerate(args.shapefiles):
        shapeopts = shapefile.split(',')
        shapepath = shapeopts[0]
        shapeoptdict = eval('dict(' + ','.join(shapeopts[1:]) + ')')
        shapename = os.path.basename(shapepath)[:-3] + str(si)
        map.readshapefile(shapepath, shapename, ax=ax, **shapeoptdict)
    args.map = map
    fig = plt.gcf()
    if len(args.figure_keywords) > 0:
        plt.setp(fig, **args.figure_keywords)

    ax = plt.gca()
    if len(args.axes_keywords) > 0:
        plt.setp(ax, **args.axes_keywords)

    map = args.map
    nborders = len(ax.collections)
    for fi, ifile in enumerate(ifiles):
        if map.projection in ('lcc', 'merc'):
            lat = ifile.variables['latitude']
            lon = ifile.variables['longitude']
            latb, latunit = getybnds(ifile)[:]
            lonb, lonunit = getxbnds(ifile)[:]
        else:
            lat = ifile.variables['latitude']
            lon = ifile.variables['longitude']
            latb, latunit = getlatbnds(ifile)[:]
            lonb, lonunit = getlonbnds(ifile)[:]

        if latb.ndim == lonb.ndim and lonb.ndim == 2:
            LON, LAT = lonb, latb
        else:
            LON, LAT = np.meshgrid(lonb.view(np.ndarray),
                                   latb.view(np.ndarray))

        variables = args.variables
        if variables is None:
            def isgeo(var):
                geo2d = set(['latitude', 'longitude'])
                vard = getattr(var, 'coordinates', '').split()
                hasgeo2d = len(geo2d.intersection(vard)) == 2
                return hasgeo2d
            variables = [key for key, var in ifile.variables.items()
                         if isgeo(var)]
        if len(variables) == 0:
            raise ValueError('Unable to heuristically determin plottable ' +
                             'variables; use -v to specify variables for ' +
                             'plotting')
        for varkey in variables:
            ax = plt.gca()

            if not args.overlay:
                del ax.collections[nborders:]
            var = ifile.variables[varkey]
            if args.squeeze:
                vals = var[:].squeeze()
            else:
                vals = var[:]
            vmin, vmax = vals.min(), vals.max()
            if args.normalize is None:
                from scipy.stats import normaltest
                if normaltest(vals.ravel())[1] < 0.001:
                    cvals = np.ma.compressed(vals)
                    boundaries = np.percentile(cvals, np.arange(0, 110, 10))
                    warn('Autoselect deciles colormap of %s; override ' +
                         'width --norm' % varkey)
                else:
                    boundaries = np.linspace(vmin, vmax, num=11)
                    warn(('Autoselect linear colormap of %s; override ' +
                          'width --norm') % varkey)
                ordermag = (boundaries.max() /
                            np.ma.masked_values(boundaries, 0).min())
                if (ordermag) > 10000:
                    formatter = LogFormatter(labelOnlyBase=False)
                else:
                    formatter = None
                norm = BoundaryNorm(boundaries, ncolors=256)
            else:
                norm = eval(args.normalize)
                formatter = None
            if args.colorbarformatter is not None:
                try:
                    formatter = eval(args.colorbarformatter)
                except Exception:
                    formatter = args.colorbarformatter

            if norm.vmin is not None:
                vmin = norm.vmin
            if norm.vmax is not None:
                vmax = norm.vmax
            varunit = getattr(var, 'units', 'unknown').strip()
            if args.verbose > 0:
                print(varkey, sep='')
            if vals.ndim == 1:
                notmasked = ~(np.ma.getmaskarray(lon[:]) |
                              np.ma.getmaskarray(lat[:]) |
                              np.ma.getmaskarray(vals[:]))
                scatlon = lon[:][notmasked]
                scatlat = lat[:][notmasked]
                scatvals = vals[:][notmasked]
                patches = map.scatter(scatlon[:], scatlat[:], c=scatvals,
                                      edgecolors='none', s=24, norm=norm,
                                      ax=ax, zorder=2)
            else:
                if vals.ndim != 2:
                    dimlendictstr = str(dict(zip(var.dimensions, var.shape)))
                    warn('Maps require 2-d data; values right now %s %s' %
                         (str(vals.shape), dimlendictstr))
                patches = map.pcolor(LON, LAT, vals, norm=norm, ax=ax)
            if lonunit == 'x (m)':
                ax.xaxis.get_major_formatter().set_scientific(True)
                ax.xaxis.get_major_formatter().set_powerlimits((-3, 3))
            if latunit == 'y (m)':
                ax.yaxis.get_major_formatter().set_scientific(True)
                ax.yaxis.get_major_formatter().set_powerlimits((-3, 3))
            ax.set_xlabel(lonunit)
            ax.set_ylabel(latunit)
            height = np.abs(np.diff(ax.get_ylim()))
            width = np.abs(np.diff(ax.get_xlim()))
            if width >= height:
                orientation = 'horizontal'
            else:
                orientation = 'vertical'
            if cbar is None:
                cax = None
            else:
                cax = cbar.ax
                cax.cla()

            if vals.max() > vmax and vals.min() < vmin:
                extend = 'both'
            elif vals.max() > vmax:
                extend = 'max'
            elif vals.min() < vmin:
                extend = 'min'
            else:
                extend = 'neither'
            cbar = plt.gcf().colorbar(patches, orientation=orientation,
                                      cax=cax, extend=extend, format=formatter,
                                      spacing='proportional')
            del cbar.ax.texts[:]
            varminmaxtxt = ('; min=%.3g; max=%.3g)' %
                            (var[:].min(), var[:].max()))
            cbar.set_label(varkey + ' (' + varunit + varminmaxtxt)
            # if orientation == 'vertical':
            #     cbar.ax.text(.5, 1.05, '%.3g' % var[:].max(),
            #                  horizontalalignment = 'center',
            #                  verticalalignment = 'bottom')
            #     cbar.ax.text(.5, -.06, '%.3g ' % var[:].min(),
            #                  horizontalalignment = 'center',
            #                  verticalalignment = 'top')
            # else:
            #     cbar.ax.text(1.05, .5, ' %.3g' % var[:].max(),
            #                  verticalalignment = 'center',
            #                  horizontalalignment = 'left')
            #     cbar.ax.text(-.06, .5, '%.3g ' % var[:].min(),
            #                  verticalalignment = 'center',
            #                  horizontalalignment = 'right')
            cbar.update_ticks()
            fmt = args.figformat
            outpath = args.outpath
            if len(ifiles) > 1:
                lstr = str(fi).rjust(len(str(len(ifiles))), '0')
                if args.verbose > 0:
                    print('adding numeric suffix for file', lstr)
            else:
                lstr = ''

            figpath = os.path.join(outpath + varkey + lstr + '.' + fmt)
            if args.interactive:
                csl = PNCConsole(locals=globals())
                csl.interact()
            for cmd in args.plotcommands:
                exec(cmd)
            plt.savefig(figpath)
            if args.verbose > 0:
                print('Saved fig', figpath)
